@@ -2,29 +2,31 @@
 
 Invocation flow:
 
-    1. Read, validate and process the input (args, `stdin`).
-    2. Create and send a request.
-    3. Stream, and possibly process and format, the requested parts
-       of the request-response exchange.
-    4. Simultaneously write to `stdout`
-    5. Exit.
+  1. Read, validate and process the input (args, `stdin`).
+  2. Create and send a request.
+  3. Stream, and possibly process and format, the parts
+     of the request-response exchange selected by output options.
+  4. Simultaneously write to `stdout`
+  5. Exit.
 
 """
 import sys
 import errno
 
 import requests
-from httpie import __version__ as httpie_version
 from requests import __version__ as requests_version
 from pygments import __version__ as pygments_version
 
-from .compat import str, is_py3
-from .client import get_response
-from .downloads import Download
-from .models import Environment
-from .output import build_output_stream, write, write_with_colors_win_py3
-from . import ExitStatus
-from .plugins import plugin_manager
+from httpie import __version__ as httpie_version, ExitStatus
+from httpie.compat import str, bytes, is_py3
+from httpie.client import get_response
+from httpie.downloads import Download
+from httpie.context import Environment
+from httpie.plugins import plugin_manager
+from httpie.output.streams import (
+    build_output_stream,
+    write, write_with_colors_win_py3
+)
 
 
 def get_exit_status(http_status, follow=False):
@@ -43,7 +45,7 @@ def get_exit_status(http_status, follow=False):
 
 
 def print_debug_info(env):
-    sys.stderr.writelines([
+    env.stderr.writelines([
         'HTTPie %s\n' % httpie_version,
         'HTTPie data: %s\n' % env.config.directory,
         'Requests %s\n' % requests_version,
@@ -52,14 +54,29 @@ def print_debug_info(env):
     ])
 
 
+def decode_args(args, stdin_encoding):
+    """
+    Convert all bytes ags to str
+    by decoding them using stdin encoding.
+
+    """
+    return [
+        arg.decode(stdin_encoding)
+        if type(arg) == bytes else arg
+        for arg in args
+    ]
+
+
 def main(args=sys.argv[1:], env=Environment()):
     """Run the main program and write the output to ``env.stdout``.
 
     Return exit status code.
 
     """
+    args = decode_args(args, env.stdin_encoding)
     plugin_manager.load_installed_plugins()
-    from .cli import parser
+
+    from httpie.cli import parser
 
     if env.config.default_options:
         args = env.config.default_options + args
@@ -146,12 +163,17 @@ def main(args=sys.argv[1:], env=Environment()):
                 env.stderr.write('\n')
             else:
                 raise
-    except (KeyboardInterrupt, SystemExit):
+    except KeyboardInterrupt:
         if traceback:
             raise
         env.stderr.write('\n')
         exit_status = ExitStatus.ERROR
-
+    except SystemExit as e:
+        if e.code != ExitStatus.OK:
+            if traceback:
+                raise
+            env.stderr.write('\n')
+            exit_status = ExitStatus.ERROR
     except requests.Timeout:
         exit_status = ExitStatus.ERROR_TIMEOUT
         error('Request timed out (%ss).', args.timeout)
