@@ -4,11 +4,10 @@
 import re
 import os
 
-import requests
 from requests.cookies import RequestsCookieJar, create_cookie
 
-from .compat import urlsplit
-from .config import BaseConfigDict, DEFAULT_CONFIG_DIR
+from httpie.compat import urlsplit
+from httpie.config import BaseConfigDict, DEFAULT_CONFIG_DIR
 from httpie.plugins import plugin_manager
 
 
@@ -21,21 +20,21 @@ VALID_SESSION_NAME_PATTERN = re.compile('^[a-zA-Z0-9_.-]+$')
 SESSION_IGNORED_HEADER_PREFIXES = ['Content-', 'If-']
 
 
-def get_response(session_name, requests_kwargs, config_dir, args,
-                 read_only=False):
+def get_response(requests_session, session_name,
+                 config_dir, args, read_only=False):
     """Like `client.get_response`, but applies permanent
     aspects of the session to the request.
 
     """
+    from .client import get_requests_kwargs, dump_request
     if os.path.sep in session_name:
         path = os.path.expanduser(session_name)
     else:
-        hostname = (
-            requests_kwargs['headers'].get('Host', None)
-            or urlsplit(requests_kwargs['url']).netloc.split('@')[-1]
-        )
-
-        assert re.match('^[a-zA-Z0-9_.:-]+$', hostname)
+        hostname = (args.headers.get('Host', None)
+                    or urlsplit(args.url).netloc.split('@')[-1])
+        if not hostname:
+            # HACK/FIXME: httpie-unixsocket's URLs have no hostname.
+            hostname = 'localhost'
 
         # host:port => host_port
         hostname = hostname.replace(':', '_')
@@ -47,9 +46,10 @@ def get_response(session_name, requests_kwargs, config_dir, args,
     session = Session(path)
     session.load()
 
-    request_headers = requests_kwargs.get('headers', {})
-    requests_kwargs['headers'] = dict(session.headers, **request_headers)
-    session.update_headers(request_headers)
+    kwargs = get_requests_kwargs(args, base_headers=session.headers)
+    if args.debug:
+        dump_request(kwargs)
+    session.update_headers(kwargs['headers'])
 
     if args.auth:
         session.auth = {
@@ -58,25 +58,24 @@ def get_response(session_name, requests_kwargs, config_dir, args,
             'password': args.auth.value,
         }
     elif session.auth:
-        requests_kwargs['auth'] = session.auth
+        kwargs['auth'] = session.auth
 
-    requests_session = requests.Session()
     requests_session.cookies = session.cookies
 
     try:
-        response = requests_session.request(**requests_kwargs)
+        response = requests_session.request(**kwargs)
     except Exception:
         raise
     else:
         # Existing sessions with `read_only=True` don't get updated.
-        if session.is_new or not read_only:
+        if session.is_new() or not read_only:
             session.cookies = requests_session.cookies
             session.save()
         return response
 
 
 class Session(BaseConfigDict):
-    helpurl = 'https://github.com/jkbr/httpie#sessions'
+    helpurl = 'https://github.com/jakubroztocil/httpie#sessions'
     about = 'HTTPie session file'
 
     def __init__(self, path, *args, **kwargs):
@@ -102,7 +101,7 @@ class Session(BaseConfigDict):
 
         """
         for name, value in request_headers.items():
-
+            value = value.decode('utf8')
             if name == 'User-Agent' and value.startswith('HTTPie/'):
                 continue
 
